"""Visualise a heatmap of two columns of discrete labels. Used to visualise results of two cell label columns."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def heatmap(
    df: pd.DataFrame,
    x: str,
    y: str,
    cols: list = [],
    normalise: str = "",
    common_cols: bool = True,
    ax: plt.Axes = None,
    log_counts: bool = False,
    **kwargs,
):
    """Create a heatmap count of the values in the dataframe.

    Args:
        df (pd.DataFrame): DataFrame containing the data to plot.
        x (str): Column name for the x-axis (rows).
        y (str): Column name for the y-axis (columns).
        cols (list, optional): Order of columns and rows. Defaults to descending counts.
        common_cols (bool, optional): If True, ensure all columns are present in box x and y. Defaults to True.
        normalise (str, optional): Normalise by "all", "row" or "col". Defaults to "" (no normalisation).
        ax (plt.Axes, optional): Axes to plot on. If None, a new figure is created. Defaults to None.
        log_counts (bool, optional): If True, log-transform the counts. Defaults to False.
        **kwargs: Additional keyword arguments for seaborn heatmap.

    """
    # params
    fmt_default = ".0f" if not log_counts else ".2f"
    kwargs_heatmap_defaults = {
        "cbar": False,
        "cmap": "Blues",
        "annot": True,
        "fmt": fmt_default,
    }
    kwargs_heatmap = {**kwargs_heatmap_defaults, **kwargs}

    # sort cols
    cols_all_ref = list(
        set(df[x].astype(str).unique()) | set(df[y].astype(str).unique())
    )
    cols_all = cols + list(set(cols_all_ref) - set(cols))

    # aggregate
    df_agg = df.groupby([x, y], observed=True).size().reset_index(name="count")
    df_pivot = df_agg.pivot(index=y, columns=x, values="count").fillna(0)
    # ensure all cols are present
    if common_cols:
        df_pivot = df_pivot.reindex(index=cols_all, columns=cols_all).fillna(0)

    # sort if no cols were given
    if len(cols) == 0:
        row_sums = df_pivot.sum(axis=1)
        rows_order = row_sums.sort_values(ascending=False).index
    else:
        rows_order = pd.Index(cols_all)
    if common_cols:
        df_pivot = df_pivot.loc[
            rows_order[rows_order.isin(df_pivot.index)],
            rows_order[rows_order.isin(df_pivot.columns)],
        ]

    # normalise
    if normalise == "row":
        df_pivot = df_pivot.div(df_pivot.sum(axis=1), axis=0)
    elif normalise == "col":
        df_pivot = df_pivot.div(df_pivot.sum(axis=0), axis=1)
    elif normalise == "all":
        df_pivot = df_pivot.div(df_pivot.sum().sum(), axis=None)

    # log
    if log_counts:
        df_pivot = df_pivot.applymap(lambda x: np.log1p(x) if x > 0 else 0)

    # plot
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))
    ax = sns.heatmap(df_pivot, ax=ax, **kwargs_heatmap)
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    return ax



"""Introducing two metrics: Singlet Accuracy (SA) and Singlet Recovery Rate (SRR)."""

import numpy as np
import pandas as pd
from sklearn.metrics import f1_score

# y_true: ground truth labels
# y_pred: predicted labels from your algorithm


def sa(
    ground_truth,
    prediction,
    non_singlet_substring: list = ["low_quality", "doublet", ":", "negative"],
    ignore_case: bool = True,
):
    """Singlet Accuracy (SA) meassures the percentage of predicted singlets that are correctly classified.

    Args:
        ground_truth (pd.Series): Ground truth labels.
        prediction (pd.Series): Predicted labels.
        non_singlet_substring (list): List of substrings that indicate a non-singlet label.
        ignore_case (bool): Whether to ignore case when checking for non-singlet substrings.

    """
    # ignore case
    if ignore_case:
        prediction = prediction.str.lower()
        non_singlet_substring = [sub.lower() for sub in non_singlet_substring]

    # subset predicted singlets
    is_predicted_singlet = np.all(
        [~prediction.str.contains(sub) for sub in non_singlet_substring], axis=0
    )
    predicted_singlets = prediction[is_predicted_singlet]
    ground_truth_singlets = ground_truth[is_predicted_singlet]

    # return accuracy
    accuracy = (predicted_singlets == ground_truth_singlets).mean()
    return accuracy


def srr(
    ground_truth,
    prediction,
    non_singlet_substring: list = ["low_quality", "doublet", ":", "negative"],
    ignore_case: bool = True,
):
    """Singlet Recovery Rate (SRR) measures the percentage of ground truth singlets that are correctly predicted.

    Args:
        ground_truth (pd.Series): Ground truth labels.
        prediction (pd.Series): Predicted labels.
        non_singlet_substring (list): List of substrings that indicate a non-singlet label.
        ignore_case (bool): Whether to ignore case when checking for non-singlet substrings.

    """
    # ignore case
    if ignore_case:
        prediction = prediction.str.lower()
        non_singlet_substring = [sub.lower() for sub in non_singlet_substring]

    # subset ground truth singlets
    is_true_singlet = np.all(
        [~ground_truth.str.contains(sub) for sub in non_singlet_substring], axis=0
    )
    predicted_singlets = prediction[is_true_singlet]
    ground_truth_singlets = ground_truth[is_true_singlet]

    # return accuracy
    recovery_rate = (predicted_singlets == ground_truth_singlets).mean()
    return recovery_rate


def get_metrics(ground_truth, prediction):
    """Obtain metrices of a ground_truth and prediction.

    - SA (Singlet Accuracy) = Correctly Labelled Predicted Singlets / All Predicted Singlets
    - SRR (Singlet Recovery Rate) = Correctly Labelled Ground Truth Singlets / All Ground Truth Singlets
    """
    gt_negatives = ground_truth.isin(["low_quality"])
    gt_doublets = ground_truth.isin(["doublet"])
    gt_singlets = ~gt_negatives & ~gt_doublets
    pred_negatives = prediction.isin(["low_quality"])
    pred_doublets = prediction.isin(["doublet"])
    pred_singlets = ~pred_negatives & ~pred_doublets

    correct = ground_truth == prediction

    # singlet recovery rate
    srr = np.sum(gt_singlets & correct) / np.sum(gt_singlets)
    # singlet accuracy
    sa = np.sum(pred_singlets & correct) / np.sum(pred_singlets)
    # f1 score
    f1 = f1_score(ground_truth, prediction, average="micro")

    metrics = {
        "SRR": srr,
        "SA": sa,
        "F1": f1,
    }

    return pd.Series(metrics)



import logging
from pprint import pformat

import anndata as ad


def _log_parameters(logger: logging.Logger, params: dict):
    """Log parameters to the logger."""
    params = {k: v.shape if isinstance(v, ad.AnnData) else v for k, v in params.items()}
    params_str = pformat(params, indent=4)
    logger.debug(f"Parameters:\n{params_str}")


# setup logging to the console in the format:
# <timestamp> <log level> - <message>
def get_logger(logger, level: int = 1):
    # Set logger level
    log = logging.getLogger(logger)
    level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(level, logging.INFO)
    log.setLevel(level)

    # Clear existing handlers
    if log.hasHandlers():
        log.handlers.clear()

    formatter = logging.Formatter(
        "%(asctime)s HTO-DND %(name)s | %(levelname)s: %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    log.addHandler(ch)

    # Add log_parameters function to the logger
    log.log_parameters = lambda params: _log_parameters(log, params)
    return log



"""Run external tool: GMM-Demux.

Source: https://github.com/CHPGenetics/GMM-Demux
"""

import os
import subprocess
import tempfile

import numpy as np
import pandas as pd

from hto._exceptions import UserInputError
from hto._logging import get_logger


def gmm_demux(
    df: pd.DataFrame,
    hash_id: str = "hash_id",
    to_counts: bool = None,
    params: str = "",
    verbose: int = 1,
    **kwargs,
) -> pd.DataFrame:
    """Run GMM-Demux on filtered HTO. GMM-Demux uses raw counts. However, the input to this function is normalised, log-transformed and denoised.

    Use exponentiation to get an approximation of the cleaned counts.

    Args:
        df (pd.DataFrame): Dataframe with HTO counts
        hash_id (str): Hash id for output
        to_counts (bool): Whether to convert logged values to counts. If none, try to infer from the data.
        params (str): Parameters for GMM demux
        verbose (int): Verbosity level
        **kwargs: Additional arguments

    """
    # assertions
    prohibited_params = ["-f ", "--full ", "-o ", "--output ", "--csv ", "-c "]
    for param in prohibited_params:
        if param in params:
            raise UserInputError(
                f"Prohibited parameter '{param}' found in params. Please remove it."
            )

    # set tmp path
    path_tmp_dir = os.environ.get(
        "DND_TMPDIR", os.path.join(os.path.expanduser("~"), ".dnd_tmp")
    )
    os.makedirs(path_tmp_dir, exist_ok=True)
    path_tmp = tempfile.TemporaryDirectory(dir=path_tmp_dir, delete=False)
    path_tmp_str = path_tmp.name
    logger = get_logger("gmm-demux", level=verbose)

    # set to_counts
    # - if integers, then False
    # - if rounded floats equal to integers, then False
    # - if floats, then True
    # - else, raise error
    if to_counts is None:
        if all([np.issubdtype(t, np.integer) for t in df.dtypes]):
            to_counts = False
        if np.all(df.iloc[:1000].round(0) == df.iloc[:1000]):
            to_counts = False
        elif all([np.issubdtype(t, np.floating) for t in df.dtypes]):
            to_counts = True
        else:
            raise UserInputError(
                f"Cannot infer 'to_counts' from data types: {df.dtypes}. Please set 'to_counts' explicitly."
            )

    # transform to counts
    if to_counts:
        df = np.exp(df) - 1

    # write dataframe
    path_df = os.path.join(path_tmp_str, "hto.csv")
    df.to_csv(path_df, index=True, header=True)

    # get hto_array
    hto_array = kwargs.get("hto_array", ",".join(df.columns))

    # run gmm
    cmd = f"cd {path_tmp_str} && GMM-demux {path_df} {hto_array} --output {path_tmp_str} --full FULL {params} --csv"
    try:
        _ = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
        logger.info("GMM demux completed successfully.")
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"GMM demux failed with return code {e.returncode}.\n"
            f"Command: '{cmd}'\n"
            f"STDOUT: {e.stdout}\n"
            f"STDERR: {e.stderr}"
        )

    # read assignment
    # - mark doublets with : instead of -
    # - note that celltypes may include -, so first replace those with <dash>, then replace - with : and finally replace <dash> with -
    df_config = pd.read_csv(
        os.path.join(path_tmp_str, "FULL/GMM_full.config"),
        index_col=None,
        sep=", ",
        header=None,
        names=["Cluster_id", hash_id],
        engine="python",
    )
    for c in df.columns:
        df_config[hash_id] = df_config[hash_id].str.replace(c, c.replace("-", "<dash>"))
    df_config[hash_id] = df_config[hash_id].str.replace("-", ":")
    df_config[hash_id] = df_config[hash_id].str.replace("<dash>", "-")

    # join assignment
    df_full = pd.read_csv(os.path.join(path_tmp_str, "FULL/GMM_full.csv"), index_col=0)
    df_out = df_full.join(
        df_config, on="Cluster_id", how="left", lsuffix="_full", rsuffix="_config"
    )

    # rename index to "barcode"
    df_out.index.name = "barcode"

    # clean up
    path_tmp.cleanup()

    return df_out[[hash_id]]



import click

# shared default values
DEFAULTS = {
    # general
    "adata_hto": None,
    "adata_hto_raw": None,
    "adata_gex": None,
    "adata_background": None,
    "adata_out": None,
    "csv_out": None,
    "path_report": None,
    "use_layer": None,
    "inplace": False,
    "verbose": 1,
    "_required_write": False,
    # normalise
    "pseudocount": 10,
    "add_key_normalise": "normalised",
    "background_quantile": 0.3,
    # denoise
    "background_method": "kmeans-fast",
    "covariates": None,
    "design": None,
    "add_key_denoised": "denoised",
    "denoise_version": "v2",
    "kwargs_denoise": {
        "C": 1,
        "epsilon": 1,
        "loss": "squared_epsilon_insensitive",
        "intercept_scaling": 1,
    },
    # demux
    "demux_method": "gmm",
    "enforce_larger_than_background": True,
    "add_key_hashid": "hash_id",
    "add_key_doublet": "doublet_info",
    "add_key_labels": None,
    "kwargs_classify": {
        "kmeans_placeholder": -1,
        "gmm-p-cutoff": 0.5,
        "otsu_placeholder": -1,
    },
    # build_background
    "min_umi": 300,
    "next_k_cells": 10000,
    "k_gex_cells": 40000,
    "background_version": "v3",
}

# shared descriptions
DESCRIPTIONS = {
    # general
    "adata_hto": "AnnData object containing unfiltered protein expression data.",
    "adata_hto_raw": "AnnData object containing raw protein expression data.",
    "adata_gex": "AnnData object containing raw gene expression data.",
    "adata_background": f"AnnData object containing background data. Default is {DEFAULTS['adata_background']}.",
    "adata_out": f"Path to save the output AnnData object. Default is {DEFAULTS['adata_out']}.",
    "csv_out": f"Path to save demultiplexing results as CSV file. Default is {DEFAULTS['csv_out']}.",
    "path_report": f"Path to save the output report. Only created if not None. Default is {DEFAULTS['path_report']}.",
    "verbose": f"Verbosity level. Default is {DEFAULTS['verbose']}.",
    "inplace": f"Whether to perform the operation in place. Default is {DEFAULTS['inplace']}.",
    "use_layer": f"Layer to use for denoising. Default is {DEFAULTS['use_layer']}.",
    "_required_write": f"Internal parameter. Writing adata is required when run through CLI. Default is {DEFAULTS['_required_write']}.",
    # normalise
    "pseudocount": f"Value to add to the counts matrix before log-transformation. Default is {DEFAULTS['pseudocount']}.",
    "add_key_normalise": f"Key to store the normalized data in the AnnData object. Default is {DEFAULTS['add_key_normalise']}.",
    "background_quantile": f"Quantile to use for background estimation. Last resort only. Default is {DEFAULTS['background_quantile']}.",
    # denoise
    "background_method": f"Method to use for background estimation. Must be either 'kmeans-fast', 'gmm' or 'kmeans'. Default is {DEFAULTS['background_method']}.",
    "add_key_denoised": f"Key to store the denoised data in the AnnData object. Default is {DEFAULTS['add_key_denoised']}.",
    "covariates": f"Matrix of covariates to use for denoising. Not recommended for general use. Default is {DEFAULTS['covariates']}.",
    "design": f"Design matrix to use for denoising. Not recommended for general use. Default is {DEFAULTS['design']}.",
    "denoise_version": f"Version of the denoising algorithm. Must be either 'v1' or 'v2'. Default is {DEFAULTS['denoise_version']}.",
    "kwargs_denoise": f"Additional parameters for the denoising algorithm. Default is {DEFAULTS['kwargs_denoise']}.",
    # demux
    "demux_method": f"Method to use for demultiplexing. Must be either 'kmeans', 'gmm' or 'otsu'. Default is {DEFAULTS['demux_method']}.",
    "enforce_larger_than_background": f"Enforce that only cells with larger than background counts are considered for a hashtag label. This ensures that normalised counts are larger than 0. Default is {DEFAULTS['enforce_larger_than_background']}.",
    "add_key_hashid": f"Column to store the demultiplexed cell type in the AnnData object. Default is {DEFAULTS['add_key_hashid']}.",
    "add_key_doublet": f"Column to store the doublet information in the AnnData object. Default is {DEFAULTS['add_key_doublet']}.",
    "add_key_labels": f"Adata layer to store the demultiplexed labels in the AnnData object. Default is {DEFAULTS['add_key_labels']}.",
    "kwargs_classify": f"Additional parameters for the demultiplexing algorithm. Default is {DEFAULTS['kwargs_classify']}.",
    # build_background
    "min_umi": f"Minimum UMI count to consider a barcode. Default is {DEFAULTS['min_umi']}.",
    "next_k_cells": f"Number of cells to add to the background. Default is {DEFAULTS['next_k_cells']}.",
    "k_gex_cells": f"Number of cells to use for GEX-based background estimation. Default is {DEFAULTS['k_gex_cells']}.",
    "background_version": f"Version of the background building algorithm. Must be either 'v1', 'v2' or 'v3'. 'v3' is recommended for best results. Default is {DEFAULTS['background_version']}.",
}

# shared click cli options
OPTIONS = {
    # general
    "adata_hto": click.option(
        "--adata-hto",
        "-f",
        type=click.Path(exists=True),
        help=DESCRIPTIONS["adata_hto"],
        required=True,
    ),
    "adata_hto_raw": click.option(
        "--adata-hto-raw",
        "-r",
        type=click.Path(exists=True),
        help=DESCRIPTIONS["adata_hto_raw"],
        required=True,
    ),
    "adata_gex": click.option(
        "--adata-gex",
        "-g",
        type=click.Path(exists=True),
        help=DESCRIPTIONS["adata_gex"],
        required=False,
    ),
    "adata_background": click.option(
        "--adata-background",
        "-b",
        type=click.Path(exists=True),
        help=DESCRIPTIONS["adata_background"],
    ),
    "adata_out": click.option(
        "--adata-out",
        "-o",
        type=click.Path(),
        help=DESCRIPTIONS["adata_out"],
        default=DEFAULTS["adata_out"],
        required=True,
    ),
    "csv_out": click.option(
        "--csv-out",
        "-c",
        type=click.Path(),
        help=DESCRIPTIONS["csv_out"],
        default=DEFAULTS["csv_out"],
    ),
    "path_report": click.option(
        "--path-report",
        "-p",
        type=click.Path(),
        help=DESCRIPTIONS["path_report"],
        default=DEFAULTS["path_report"],
    ),
    "verbose": click.option(
        "-v",
        "--verbose",
        type=int,
        default=DEFAULTS["verbose"],
        help=DESCRIPTIONS["verbose"],
    ),
    "inplace": click.option(
        "--inplace",
        is_flag=True,
        default=DEFAULTS["inplace"],
        help=DESCRIPTIONS["inplace"],
    ),
    "use_layer": click.option(
        "--use-layer",
        type=str,
        default=DEFAULTS["use_layer"],
        help=DESCRIPTIONS["use_layer"],
    ),
    # normalise
    "pseudocount": click.option(
        "--pseudocount",
        type=int,
        default=DEFAULTS["pseudocount"],
        help=DESCRIPTIONS["pseudocount"],
    ),
    "add_key_normalise": click.option(
        "--add-key-normalise",
        type=str,
        default="normalised",
        help=DESCRIPTIONS["add_key_normalise"],
    ),  # (!) <-- changed from DEFAULTS
    # denoise
    "background_method": click.option(
        "--background-method",
        type=str,
        default=DEFAULTS["background_method"],
        help=DESCRIPTIONS["background_method"],
    ),
    "covariates": click.option(
        "--covariates", type=click.Path(exists=True), help="NOT YET SUPPORT IN CLI"
    ),
    "design": click.option(
        "--design", type=click.Path(exists=True), help="NOT YET SUPPORT IN CLI"
    ),
    "add_key_denoised": click.option(
        "--add-key-denoise",
        type=str,
        default="denoised",
        help=DESCRIPTIONS["add_key_denoised"],
    ),  # (!) <-- changed from DEFAULTS
    "denoise_version": click.option(
        "--denoise-version",
        type=str,
        default=DEFAULTS["denoise_version"],
        help=DESCRIPTIONS["denoise_version"],
    ),
    # kwargs_denoise - dictionary cli not yet supported.
    # demux
    "demux_method": click.option(
        "--demux-method",
        type=str,
        default=DEFAULTS["demux_method"],
        help=DESCRIPTIONS["demux_method"],
    ),
    "enforce_larger_than_background": click.option(
        "--enforce-larger-than-background",
        is_flag=True,
        default=DEFAULTS["enforce_larger_than_background"],
        help=DESCRIPTIONS["enforce_larger_than_background"],
    ),
    "add_key_hashid": click.option(
        "--add-key-hashid",
        type=str,
        default=DEFAULTS["add_key_hashid"],
        help=DESCRIPTIONS["add_key_hashid"],
    ),
    "add_key_doublet": click.option(
        "--add-key-doublet",
        type=str,
        default=DEFAULTS["add_key_doublet"],
        help=DESCRIPTIONS["add_key_doublet"],
    ),
    "add_key_labels": click.option(
        "--add-key-labels",
        type=str,
        default=DEFAULTS["add_key_labels"],
        help=DESCRIPTIONS["add_key_labels"],
    ),
    "kwargs_classify": click.option(
        "--kwargs-classify",
        type=(str, float),
        multiple=True,
        help="Additional parameters for the demultiplexing algorithm. Use key-value pairs, e.g. --kwargs-classify gmm-p-cutoff 0.8.",
    ),
    # build_background
    "min_umi": click.option(
        "--min-umi", type=int, default=DEFAULTS["min_umi"], help=DESCRIPTIONS["min_umi"]
    ),
    "next_k_cells": click.option(
        "--next-k-cells",
        type=int,
        default=DEFAULTS["next_k_cells"],
        help=DESCRIPTIONS["next_k_cells"],
    ),
    "k_gex_cells": click.option(
        "--k-gex-cells",
        type=int,
        default=DEFAULTS["k_gex_cells"],
        help=DESCRIPTIONS["k_gex_cells"],
    ),
    "background_version": click.option(
        "--background-version",
        type=str,
        default=DEFAULTS["background_version"],
        help=DESCRIPTIONS["background_version"],
    ),
}



"""Construct a set of cells that are used as a background reference for HTO data."""

import anndata as ad
import numpy as np

from .._defaults import DEFAULTS
from .._exceptions import AnnDataFormatError, UserInputError
from .._logging import get_logger
from .._utils import _assert_required_inputs, add_docstring, get_layer, subset_whitelist

_background_version_meta = {
    "v1": {
        "required": ["adata_hto_raw", "adata_gex"],
        "optional": ["min_umi"],
        "description": "Get a whitelist based on raw GEX counts, cutoff by min_umi. Recommend manual fine-tuning of 'min_umi' parameter.",
    },
    "v2": {
        "required": ["adata_hto", "adata_hto_raw"],
        "optional": ["next_k_cells"],
        "description": "Get the next k largest cells from each HTO of the raw HTO data that are not in the current HTO data. Not recommended.",
    },
    "v3": {
        "required": ["adata_hto", "adata_hto_raw", "adata_gex"],
        "optional": ["k_gex_cells"],
        "description": "Choose the k cells with the highest total counts from the GEX data that are not whitelisted. Recommended.",
    },
}


def _assert_background(adata, _run_assert=True):
    n = adata.shape[0]
    if not _run_assert:
        return
    elif n == 0:
        raise AnnDataFormatError("No background barcodes found in HTO data.")
    elif n < 100:
        raise AnnDataFormatError(f"Only {n} barcodes found in HTO data.")


def _log_background(n_background, n_empty, logger):
    msg = (
        f"Building set of background barcodes. "
        f"# Background: {n_background} | "
        f"# Empty: {n_empty} | "
    )
    logger.info(msg)


@add_docstring()
def build_background(
    background_version: str = DEFAULTS["background_version"],
    _run_assert: bool = True,
    **kwargs,
):
    """Build a background whitelisted adata for HTO data.

    Args:
        background_version (str): {background_version}
        **kwargs: Additional keyword arguments for the specific version

    Returns:
        AnnData: Filtered AnnData object of background data

    """
    # don't do anything if adata_background is provided
    adata_background = kwargs.get("adata_background", None)

    # assert inputs
    _assert_required_inputs(
        _background_version_meta, background_version, kwargs, "background_version"
    )
    params_required = {
        k: kwargs[k] for k in _background_version_meta[background_version]["required"]
    }
    params_optional = {
        k: kwargs.get(k, DEFAULTS[k])
        for k in _background_version_meta[background_version]["optional"]
    }
    if adata_background is not None:
        adata_background = adata_background
    elif background_version == "v1":
        adata_background = build_background_v1(
            **params_required,
            **params_optional,
            verbose=kwargs.get("verbose", DEFAULTS["verbose"]),
        )
    elif background_version == "v2":
        adata_background = build_background_v2(
            **params_required,
            **params_optional,
            verbose=kwargs.get("verbose", DEFAULTS["verbose"]),
        )
    elif background_version == "v3":
        adata_background = build_background_v3(
            **params_required,
            **params_optional,
            verbose=kwargs.get("verbose", DEFAULTS["verbose"]),
        )
    else:
        raise ValueError(
            f"Invalid version: {background_version}. Must be 'v1', 'v2' or 'v3'."
        )

    _assert_background(adata_background, _run_assert=_run_assert)
    return adata_background


@add_docstring()
def build_background_v1(
    adata_hto_raw: ad.AnnData,
    adata_gex: ad.AnnData,
    use_layer: str = DEFAULTS["use_layer"],
    min_umi: int = DEFAULTS["min_umi"],
    verbose: int = DEFAULTS["verbose"],
):
    """Get a whitelist based on GEX counts.

    Args:
        adata_hto_raw (AnnData): {adata_hto_raw}
        adata_gex (AnnData): {adata_gex}
        use_layer (str, optional): {use_layer}
        min_umi (int, optional): {min_umi}
        verbose (int, optional): {verbose}

    """
    # assertions - gex inputs must be integers
    adata_gex, x = get_layer(
        adata_gex,
        use_layer=use_layer,
        numpy=False,
        inplace=False,
        integer=True,
    )
    logger = get_logger("utils", level=verbose)

    # get whitelist
    counts = np.asarray(adata_gex.X.sum(axis=1)).flatten()
    ids_selected = set(adata_gex.obs_names[counts > min_umi])

    # subset to intersection
    ids_background = ids_selected.intersection(set(adata_hto_raw.obs_names))

    # logs
    n_background = len(ids_background)
    n_empty = adata_hto_raw.shape[0] - n_background
    _log_background(n_background, n_empty, logger)

    return adata_hto_raw[list(ids_background)]


@add_docstring()
def build_background_v2(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData,
    use_layer: str = DEFAULTS["use_layer"],
    next_k_cells: int = DEFAULTS["next_k_cells"],
    verbose: int = DEFAULTS["verbose"],
):
    """Build background by choosing the next k largest cells from the raw HTO data that are not in the current HTO data.

    Args:
        adata_hto (AnnData): {adata_hto}
        adata_hto_raw (AnnData): {adata_hto_raw}
        use_layer (str, optional): {use_layer}
        next_k_cells (int, optional): {next_k_cells}
        verbose (int, optional): {verbose}

    """
    logger = get_logger("utils", level=verbose)

    # assert that all filtered cells are in the raw data
    if not all(adata_hto.obs_names.isin(adata_hto_raw.obs_names)):
        raise UserInputError(
            "Filtered HTO data (`adata_hto`) contains cells that are not in the raw HTO data (`adata_hto_raw`). "
            "Please make sure that the filtered HTO data is a subset of the raw HTO data."
        )

    # get data
    whitelist = adata_hto.obs_names
    exclude = adata_hto_raw.obs_names.isin(whitelist)
    adata_empty = adata_hto_raw[~exclude]
    _, x = get_layer(
        adata_empty,
        use_layer=use_layer,
        numpy=True,
        inplace=False,
    )
    # get top k from each column
    background = set(whitelist)
    next_k_cells = min(next_k_cells, x.shape[0] - 1)
    if next_k_cells <= 0:
        raise UserInputError(
            "All cell-ids in the raw HTO (`adata_hto_raw`) data are already in the filtered HTO (`adata_hto`) data. "
            "Make sure to set 'next_k_cells' to a value larger than 0 and that the raw HTO data contains cells that are not in the filtered HTO data."
        )
    for i in range(x.shape[1]):
        # find value such that there are 'next_k_cells' cells larger than it
        cutoff = np.sort(x[:, i])[::-1][next_k_cells]
        sub = np.array(x[:, i] > cutoff).flatten()
        # get indices of top k cells
        sub = np.where(sub)[0][:next_k_cells]
        add_cells = set(adata_empty.obs_names[sub])
        background = background.union(add_cells)

    # logs
    n_background = len(background)
    n_empty = adata_hto_raw.shape[0] - n_background
    _log_background(n_background, n_empty, logger)

    return adata_hto_raw[list(background)]


@add_docstring()
def build_background_v3(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData,
    adata_gex: ad.AnnData,
    k_gex_cells: int = DEFAULTS["k_gex_cells"],
    use_layer: str = DEFAULTS["use_layer"],
    verbose: int = DEFAULTS["verbose"],
):
    """Choose the k cells with the highest total counts from the GEX data that are not whitelisted.

    Args:
        adata_hto (AnnData): {adata_hto}
        adata_hto_raw (AnnData): {adata_hto_raw}
        adata_gex (AnnData): {adata_gex}
        k_gex_cells (int, optional): {k_gex_cells}
        use_layer (str, optional): {use_layer}
        verbose (int, optional): {verbose}

    """
    logger = get_logger("utils", level=verbose)

    # get gex_counts
    logger.debug("Getting GEX counts...")
    adata_gex, _ = get_layer(
        adata_gex,
        use_layer=use_layer,
        numpy=False,
        inplace=False,
        integer=True,
    )
    adata_gex.obs.loc[:, "counts"] = np.asarray(adata_gex.X.sum(axis=1)).flatten()

    # filter: not cells, but in hto_raw
    df_temp = adata_gex.obs
    df_temp = df_temp[~df_temp.index.isin(adata_hto.obs_names)]
    df_temp = df_temp[df_temp.index.isin(adata_hto_raw.obs_names)]

    # select top k cells
    k_gex_cells = min(k_gex_cells, df_temp.shape[0])
    top_k_cells = df_temp.nlargest(k_gex_cells, "counts")
    whitelist = list(set(adata_hto.obs_names).union(set(top_k_cells.index)))

    # subset
    adata_background = subset_whitelist(adata_hto_raw, whitelist)

    # log
    n_background = adata_background.shape[0]
    n_empty = adata_hto_raw.shape[0] - n_background
    _log_background(n_background, n_empty, logger)

    return adata_background



"""Normalise HTO expression data based on a selected reference set of cells."""

from pprint import pformat

import anndata as ad
import numpy as np

from ._defaults import DEFAULTS
from ._exceptions import AnnDataFormatError, UserInputError
from ._logging import get_logger
from ._meta import add_meta, init_meta
from ._utils import add_docstring, get_layer
from .tl import build_background


def assert_normalisation(df, logger, max_spread=1.5, qs=[0.1, 0.99]):
    """Assert that normalised values are within a certain spread."""
    spans = []
    # get spread of each column
    for c in df.columns:
        q = df[c].quantile(qs).values
        diff = float(q[1] - q[0])
        spans.append(diff)
    ratio = max(spans) / min(spans)
    if ratio > max_spread:
        logger.warning(
            f"Spread of normalised values is too high: {ratio:.2f}. This may cause issues when estimate cell-by-cell covariates."
        )


@add_docstring()
def normalise(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData = None,
    adata_gex: ad.AnnData = None,
    adata_background: ad.AnnData = None,
    pseudocount: int = DEFAULTS["pseudocount"],
    background_version: str = DEFAULTS["background_version"],
    add_key_normalise: str = DEFAULTS["add_key_normalise"],
    use_layer: str = DEFAULTS["use_layer"],
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
    **kwargs_background,
) -> ad.AnnData:
    """Background aware HTO normalisation.

    This function implements an adapted version of the DSB algorithm (see citation), which normalizes protein
    expression data using empty droplets as background reference and optionally performs
    technical noise removal.

    Args:
        adata_hto (AnnData): {adata_hto}
        adata_hto_raw (AnnData): {adata_hto_raw}
        adata_gex (AnnData, optional): {adata_gex}
        adata_background (AnnData, optional): {adata_background}
        pseudocount (int, optional): {pseudocount}
        background_version (str, optional): {background_version}
        background_method (str, optional): {background_method}
        add_key_normalise (str, optional): {add_key_normalise}
        use_layer (str, optional): {use_layer}
        inplace (bool, optional): {inplace}
        verbose (int, optional): {verbose}
        **kwargs_background: Additional keyword arguments for the background model.

    Returns:
        ad.AnnData: AnnData object with normalized protein expression data, either
        in the original matrix or in a specified layer.

    Citation:
    Mulè, M.P., Martins, A.J. & Tsang, J.S. Normalizing and denoising protein expression data from droplet-based single cell profiling. Nat Commun 13, 2099 (2022). https://doi.org/10.1038/s41467-022-29356-8

    """
    # Get logger
    logger = get_logger("normalise", level=verbose)
    logger.log_parameters(locals())
    logger.debug("Starting normalization...")

    # Subset to same columns in hto and hto_raw
    columns_mismatch = list(
        set(adata_hto.var_names).difference(set(adata_hto_raw.var_names))
    )
    if not all(adata_hto.var_names.isin(adata_hto_raw.var_names)):
        raise UserInputError(
            f"Columns don't match. 'adata_hto' has columns that are not in 'adata_hto_raw': {columns_mismatch}. Make sure that all var_names in 'adata_hto' are also in 'adata_hto_raw'."
        )
    adata_hto_raw = adata_hto_raw[:, adata_hto.var_names].copy()

    # Get background if not provided
    adata_background = build_background(
        background_version=background_version,
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        adata_background=adata_background,
        verbose=verbose,
        **kwargs_background,
    )

    # Setup
    adata_hto, adt = get_layer(
        adata_hto, use_layer=use_layer, integer=True, numpy=True, inplace=inplace
    )

    adata_background, adtu = get_layer(
        adata_background, use_layer=use_layer, integer=True, numpy=True, inplace=inplace
    )

    # Init metadata
    adata_hto = init_meta(adata_hto)

    # Subsets
    barcodes_filtered = set(adata_hto.obs_names)
    barcodes_background = set(adata_background.obs_names)
    barcodes_background_only = barcodes_background.difference(barcodes_filtered)
    overlap_barcode = list(
        barcodes_filtered & barcodes_background
    )  # check that naming is consistent

    n_filtered = len(barcodes_filtered)
    n_background = len(barcodes_background)
    pct_background = n_filtered / n_background * 100

    logger.debug(
        f"Filtered adata: {n_filtered / 1000:.1f}K cells | Background adata: {n_background / 1000:.1f}K cells"
    )
    logger.debug(
        f"Background cells: {n_background / 1000:.1f}K cells | Overlapping cells: {len(overlap_barcode) / 1000:.1f}K cells"
    )
    if pct_background < 10:
        logger.warning(
            f"Only few barcodes are used for normalization: {n_background / 1000:.1f}K ({pct_background:.1f}%)"
        )

    # Identify barcodes that are in adata_raw but not in adata_filtered
    if len(barcodes_background_only) < 5:
        raise AnnDataFormatError("adata_raw_missing_cells", barcodes_background)
    if len(overlap_barcode) < 5:
        raise AnnDataFormatError("adata_no_overlapping_names", len(barcodes_filtered))

    # Log transform both matrices
    adt_log = np.log(adt + pseudocount)
    adtu_log = np.log(adtu + pseudocount)

    # Calculate mean and sd of log-transformed empty droplets for each protein
    mu_empty = np.array(np.mean(adtu_log, axis=0)).flatten()
    sd_empty = np.array(np.std(adtu_log, axis=0)).flatten()

    # Normalize the cell protein matrix
    normalized_matrix = (adt_log - mu_empty) / sd_empty

    # Ensure that the normalized matrix is a numpy array
    normalized_matrix = np.array(normalized_matrix)

    # Checkpoint
    if add_key_normalise is not None:
        adata_hto.layers[add_key_normalise] = normalized_matrix
        logger.info(f"Normalized matrix stored in adata.layers['{add_key_normalise}']")
    else:
        adata_hto.X = normalized_matrix
        logger.info("Normalization completed and stored in adata.X")

    # Assert
    assert_normalisation(adata_hto.to_df(add_key_normalise), logger)

    # Log metadata
    logger.debug(pformat(adata_hto.uns["dnd"]))

    # Store meta information
    adata_hto = add_meta(
        adata_hto,
        step="normalise",
        layer=add_key_normalise,
        params={
            "pseudocount": pseudocount,
            "background": adata_background.obs_names.values,
        },
        mu_empty=mu_empty,
        sd_empty=sd_empty,
    )

    return adata_hto


def normalise_debug(
    adata_hto: ad.AnnData,
    background_quantile: float = DEFAULTS["background_quantile"],
    use_layer: str = DEFAULTS["use_layer"],
    add_key_normalise: str = DEFAULTS["add_key_normalise"],
    verbose: int = DEFAULTS["verbose"],
):
    """Debug normaliser that only requires filtered HTO data.

    If no background data or GEX data is available, align the quantiles of the filtered HTO data from 0 to 1. This is clearly not recommended, but is provided as a last
    resort option.

    Args:
        adata_hto (AnnData): {adata_hto}
        background_quantile (float, tuple, optional): {background_quantile}
        use_layer (str, optional): {use_layer}
        add_key_normalise (str, optional): {add_key_normalise}
        verbose (int, optional): {verbose}

    """
    logger = get_logger("utils", level=verbose)

    # estimate empty
    adata_hto = init_meta(adata_hto)
    df = adata_hto.to_df(use_layer)
    adt_log = np.log1p(df).values

    mu_empty = []
    sd_empty = []
    for i in range(adt_log.shape[1]):
        q = np.quantile(adt_log[:, i], background_quantile)
        mu_temp = np.mean(adt_log[adt_log[:, i] < q, i])
        sd_temp = np.std(adt_log[adt_log[:, i] < q, i])
        mu_empty.append(mu_temp)
        sd_empty.append(sd_temp)
        if sd_empty[i] <= 0:
            logger.warning(
                f"Standard deviation for column {i} is zero or negative. This may cause issues with normalization. Setting it to 0.01"
            )
            sd_empty[i] = 0.01

    # normalise
    normalized_matrix = (adt_log - mu_empty) / sd_empty
    adata_hto.layers["normalised"] = normalized_matrix

    # Checkpoint
    if add_key_normalise is not None:
        adata_hto.layers[add_key_normalise] = normalized_matrix
        logger.info(f"Normalized matrix stored in adata.layers['{add_key_normalise}']")
    else:
        adata_hto.X = normalized_matrix
        logger.info("Normalization completed and stored in adata.X")

    # Assert
    assert_normalisation(adata_hto.to_df(add_key_normalise), logger)

    # Log metadata
    logger.debug(pformat(adata_hto.uns["dnd"]))

    # Store meta information
    adata_hto = add_meta(
        adata_hto,
        step="normalise",
        layer=add_key_normalise,
        params={
            "background_quantile": background_quantile,
        },
        mu_empty=mu_empty,
        sd_empty=sd_empty,
    )

    return adata_hto



from itertools import chain
from math import floor

import numpy as np
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture

from hto._logging import get_logger
from hto._utils import _assert_required_inputs

SUPPORTED_BACKGROUND_METHODS = ["kmeans-fast", "kmeans", "gmm"]

_background_method_meta = {
    "kmeans-fast": {
        "description": "Simple heuristics for fast, row-wise KMeans with 2 clusters in pandas DataFrame. Recommended",
        "required": [],
        "optional": ["n_iter", "inits"],
    },
    "kmeans": {
        "description": "Fit a KMeans model to the input data and return the mean of the first cluster.",
        "required": [],
        "optional": [],
    },
    "gmm": {
        "description": "Fit a Gaussian Mixture Model to the input data and return the mean of the first component.",
        "required": [],
        "optional": ["kwargs_denoise"],
    },
}


# ASSERTIONS
def assert_background(method, **kwargs):
    assert method in SUPPORTED_BACKGROUND_METHODS, (
        f"Method '{method}' not supported. Choose from: {', '.join(SUPPORTED_BACKGROUND_METHODS)}"
    )


# OUT OF THE BOX
def _apply_gmm_to_row(x, func):
    """Apply a function to each row of the input matrix using Gaussian Mixture Models."""
    metas = []
    n_cells = x.shape[0]
    results = [func(x[i, :]) for i in range(n_cells)]
    noise_vector = np.array([r[0] for r in results])
    metas = [r[1] for r in results]
    meta = {"debug": {"metas": metas}}
    return noise_vector, meta


def _get_background_gmm(normalized_matrix, **kwargs):
    """Fit a Gaussian Mixture Model to the input data and return the mean of the first component."""

    def _get_background(x):
        gmm = GaussianMixture(n_components=2, random_state=0, **kwargs).fit(
            x.reshape(-1, 1)
        )
        return min(gmm.means_)[0], {"debug": {"model": gmm}}

    return _apply_gmm_to_row(normalized_matrix, _get_background)


def _get_background_kmeans(normalized_matrix, **kwargs):
    """Fit a KMeans model to the input data and return the mean of the first cluster."""

    def _get_background(x):
        kmeans = KMeans(n_clusters=2, random_state=0, **kwargs).fit(x.reshape(-1, 1))
        return min(kmeans.cluster_centers_)[0], {"model": kmeans}

    return _apply_gmm_to_row(normalized_matrix, _get_background)


# CUSTOM KMEANS
def _converge(matrix, center_lower, center_upper, n_iter):
    """Converge to the background noise vectors."""
    # assert dimensions
    assert matrix.ndim == 2, "Matrix must be a 2D numpy array."
    assert isinstance(matrix, np.ndarray), "Matrix must be a numpy array."
    assert center_lower.shape == (matrix.shape[0],), (
        f"Center lower must have shape ({matrix.shape[0]}, ), got {center_lower.shape}."
    )
    assert center_upper.shape == (matrix.shape[0],), (
        f"Center upper must have shape ({matrix.shape[0]}, ), got {center_upper.shape}."
    )

    # converge
    for _ in range(n_iter):
        row_mid = (center_lower + center_upper) / 2
        assignment = matrix > row_mid[:, np.newaxis]
        center_lower = np.sum(np.where(~assignment, matrix, 0), axis=1) / np.sum(
            ~assignment, axis=1
        )
        center_upper = np.sum(np.where(assignment, matrix, 0), axis=1) / np.sum(
            assignment, axis=1
        )

    # inertia
    inertia = np.sum(
        np.where(~assignment, np.array(matrix - center_lower[:, np.newaxis]) ** 2, 0),
        axis=1,
    )
    inertia += np.sum(
        np.where(assignment, np.array(matrix - center_upper[:, np.newaxis]) ** 2, 0),
        axis=1,
    )

    # assertions (numpy updates change specific behavious)
    assert center_lower.shape == (matrix.shape[0],), (
        f"Center lower must have shape ({matrix.shape[0]},), got {center_lower.shape}."
    )
    assert center_upper.shape == (matrix.shape[0],), (
        f"Center upper must have shape ({matrix.shape[0]},), got {center_upper.shape}."
    )
    assert inertia.shape == (matrix.shape[0],), (
        f"Inertia must have shape ({(matrix.shape[0],)},), got {inertia.shape}."
    )
    return center_lower, center_upper, inertia


def _init_strategies(normalized_matrix, method, *args, **kwargs):
    """Return iterator of initialisation strategies.

    Supported methods:
    - 'rank': rank-based initialisation. args: k (int) - which rank to use from top and bottom.
    - 'rank_lower': rank-based initialisation. args: k (int) - which rank to use from bottom.
    - '1vall': Use max as top and mean of the rest as bottom.
    """
    if method == "rank":
        k = kwargs.get("k", 0)
        assert k >= 0, "Rank must be greater than 0."
        assert k < floor(normalized_matrix.shape[1] / 2), (
            "Rank must be less than half the number of HTOs."
        )
        center_lower_init = np.partition(normalized_matrix, k, axis=1)[:, k]
        center_upper_init = np.partition(normalized_matrix, -k, axis=1)[:, -k]
        yield center_lower_init, center_upper_init

    elif method == "rank_lower":
        k = kwargs.get("k", 0)
        assert k >= 0, "Rank must be greater or equal to 0."
        assert k < normalized_matrix.shape[1] - 1, (
            "Rank must be less than the number of HTOs - 1."
        )
        center_lower_init = np.partition(normalized_matrix, k, axis=1)[:, k]
        center_upper_init = np.max(normalized_matrix, axis=1)
        yield center_lower_init, center_upper_init

    elif method == "1vall":
        center_upper_init = np.max(normalized_matrix, axis=1)
        center_lower_init = np.mean(
            np.partition(normalized_matrix, -1, axis=1)[:, :-1], axis=1
        )
        yield center_lower_init, center_upper_init

    else:
        raise ValueError(f"Method '{method}' not supported.")


def _all_inits(matrix):
    """Generate all supported inits - feasible for lower dimensions."""
    from math import floor

    m = matrix.shape[1]
    inits = [{"method": "1vall"}]
    for i in range(m - 1):
        inits.append({"method": "rank_lower", "k": i})
    for i in range(floor(m / 2)):
        inits.append({"method": "rank", "k": i})
    return inits


def _get_background_kmeans_fast(matrix, n_iter=5, inits=None):
    """Apply simple heuristics for fast, row-wise KMeans with 2 clusters in pandas DataFrame.

    1. Initialize clusters using 'inits' strategies.
    2. Iterate through initialisations:
        a. Converge with simple KMeans algorithm.
        b. Update cluster centers using cluster with lowest inertia.

    Note that '1vall' is computationally efficient, while preserving almost exact results
    compared to 'KMeans' and 'GMM' for HTO data.

    Example:
    ```
    matrix = np.random.normal(0, 1, (100, 10))
    background = _get_background_kmeans_fast(matrix)
    ```

    """
    # debug logger
    logger = get_logger("hto._cluster_background", level=1)

    # init
    assert isinstance(matrix, np.ndarray), "Matrix must be a numpy array."
    if inits is None:
        inits = _all_inits(matrix)
    intertia_min = np.repeat(np.inf, matrix.shape[0])
    center_lower_min = np.zeros(matrix.shape[0])
    center_upper_min = np.zeros(matrix.shape[0])

    # get iterator
    inits = chain(*[_init_strategies(matrix, **init) for init in inits])

    for center_inits in inits:
        # converge
        center_lower, center_upper, inertia = _converge(
            matrix, *center_inits, n_iter=n_iter
        )
        # update if better
        center_lower_min = np.where(
            inertia < intertia_min, center_lower, center_lower_min
        )
        center_upper_min = np.where(
            inertia < intertia_min, center_upper, center_upper_min
        )
        intertia_min = np.minimum(inertia, intertia_min)
        # debug
        logger.debug(
            "inertia",
            inertia.shape,
            "inertia_min",
            intertia_min.shape,
            "center_lower",
            center_lower.shape,
            "center_lower_min",
            center_lower_min.shape,
            "inertia_min type",
            type(intertia_min),
            "center_lower_min type",
            type(center_lower_min),
        )

    # store data
    assignment = matrix > ((center_lower_min + center_upper_min) / 2).reshape(-1, 1)
    meta = {
        "debug": {
            "intertia": intertia_min,
            "center_lower": center_lower_min,
            "center_upper": center_upper_min,
            "assignment": assignment,
        }
    }

    # return mean of the lower cluster
    return center_lower_min, meta


def estimate_background(
    matrix,
    method="kmeans-fast",
    **kwargs,
):
    # assert inputs
    _assert_required_inputs(
        meta=_background_method_meta,
        key=method,
        kwargs=kwargs,
        parameter="background_method",
    )

    # get background
    if method == "kmeans-fast":
        return _get_background_kmeans_fast(matrix, **kwargs)
    elif method == "kmeans":
        return _get_background_kmeans(matrix, **kwargs)
    elif method == "gmm":
        return _get_background_gmm(matrix, **kwargs)
    else:
        raise ValueError(f"Method '{method}' not supported.")



"""Module containing functions for clustering and demultiplexing HTO data."""

import numpy as np
import scipy.stats
from scipy.optimize import root_scalar
from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score, silhouette_score
from sklearn.mixture import GaussianMixture

from ._defaults import DEFAULTS
from ._logging import get_logger
from ._utils import add_docstring

SUPPORTED_DEMUX_METHODS = ["kmeans", "gmm", "otsu", "gmm_demux"]


def assert_demux(method):
    if method not in SUPPORTED_DEMUX_METHODS:
        raise ValueError(
            f"Method '{method}' not supported. Choose from: {', '.join(SUPPORTED_DEMUX_METHODS)}"
        )


@add_docstring()
def classify(
    data,
    demux_method: str = DEFAULTS["demux_method"],
    verbose: int = DEFAULTS["verbose"],
    **kwargs_classify,
):
    """Perform clustering to identify positive populations for each HTO.

    Methods:
        - K-means: Silhouette score and Davies-Bouldin index
        - GMM: BIC and log-likelihood
        - Otsu: Inter-class variance and entropy
        - GMM demux: Uses external gmm_demux function

    Args:
        data (np.array): The input data used for clustering
        demux_method (str): {demux_method}
        verbose (int): {verbose}
        **kwargs_classify: {kwargs_classify}

    Returns:
        tuple: A tuple containing:
            - np.array: The cluster labels
            - float: The threshold value (or an approximation thereof)used for classification
            - dict: A dictionary containing various goodness of fit metrics

    """
    logger = get_logger("demux", level=verbose)
    if demux_method == "kmeans":
        logger.debug("Applying K-means to each column")
        return classify_by_column(_classify_kmeans_one, data, logger, **kwargs_classify)
    elif demux_method == "gmm":
        logger.debug("Applying GMM to each column")
        return classify_by_column(_classify_gmm_one, data, logger, **kwargs_classify)
    elif demux_method == "otsu":
        logger.debug("Applying Otsu's method to each column")
        return classify_by_column(_classify_otsu_one, data, logger, **kwargs_classify)
    elif demux_method == "gmm_demux":
        logger.debug("Applying GMM demux to each column")
        return classify_gmm_demux(data, logger, **kwargs_classify)
    else:
        raise ValueError(
            f"Method '{demux_method}' is not supported. Must be one of {SUPPORTED_DEMUX_METHODS}"
        )


def classify_by_column(func, df_umi, logger=None, **kwargs):
    """Demultiplex HTO data by clustering each column. Reusable for different methods."""
    # init
    classifications = {}
    metrics = {}
    thresholds = {}
    if logger is None:
        logger = get_logger("demux", level=1)

    for hto in df_umi.columns:
        # prepare data
        logger.debug(f"Demultiplexing HTO '{hto}'...")
        data = df_umi[hto].values.reshape(-1, 1)
        # apply function
        labels, threshold, hto_metrics = func(data, logger)
        # store results
        thresholds[hto] = float(threshold)
        metrics[hto] = hto_metrics
        classifications[hto] = labels

    return classifications, thresholds, metrics


def _classify_kmeans_one(series, logger=None, **kwargs):
    # predict clusters
    logger = logger or get_logger("demux", level=1)
    logger.debug("Clustering HTO data using K-means")
    model = KMeans(n_clusters=2, random_state=42, n_init=10)
    cluster = model.fit_predict(series)
    positive_cluster = np.argmax(model.cluster_centers_)
    labels = (cluster == positive_cluster).astype(int)

    # evaluate
    logger.debug("Evaluating K-means clustering")
    threshold = np.mean(model.cluster_centers_)
    silhouette = silhouette_score(series, labels)
    davies_bouldin = davies_bouldin_score(series, labels)
    metrics = {
        "silhouette_score": float(silhouette),
        "davies_bouldin_index": float(davies_bouldin),
    }
    return labels, threshold, metrics


def _classify_gmm_one(series, logger=None, **kwargs):
    # get params
    gmm_p_cutoff = kwargs.get(
        "gmm-p-cutoff", DEFAULTS["kwargs_classify"]["gmm-p-cutoff"]
    )

    # predict clusters
    logger = logger or get_logger("demux", level=1)
    logger.debug("Clustering HTO data using GMM")

    # fit model
    model = GaussianMixture(n_components=2, random_state=42)
    model.fit(series)

    # custom predict
    def custom_predict(model, series):
        """Implement custom prediction logic.

        - If x > larger means, return 1
        - If x < smaller means, return 0
        - If x is between the means, return 0 or 1 based on the posterior probability
        """
        # prepare and predict
        series = np.array(series).reshape(-1, 1)
        probabilities = model.predict_proba(series)
        larger_mean = np.argmax(model.means_)
        smaller_mean = np.argmin(model.means_)

        # update labels
        labels = probabilities[:, larger_mean] > gmm_p_cutoff
        labels[(series < model.means_[smaller_mean][0]).flatten()] = 0
        labels[(series >= model.means_[larger_mean][0]).flatten()] = 1

        return labels

    # get labels
    labels = custom_predict(model, series)

    # search threshold
    logger.debug(
        f"Calculating threshold for posterior probability cutoff {gmm_p_cutoff}"
    )

    def posterior_diff(x):
        """Optimise function."""
        return custom_predict(model, x) - gmm_p_cutoff

    # find threshold
    x_min = series.min()
    x_max = series.max()
    threshold = root_scalar(
        posterior_diff, bracket=[x_min, x_max], method="brentq"
    ).root

    # evaluate
    logger.debug("Evaluating GMM clustering")
    bic = model.bic(series)
    log_likelihood = model.score(series) * series.shape[0]
    metrics = {"bic": float(bic), "log_likelihood": float(log_likelihood)}

    return labels, threshold, metrics


def _classify_otsu_one(series, logger=None, **kwargs):
    # import
    try:
        from skimage.filters import threshold_otsu
    except ImportError:
        raise ImportError(
            "scikit-image is not installed. Install with `pip install scikit-image`"
        )

    # predict
    logger = logger or get_logger("demux", level=1)
    logger.debug("Thresholding HTO data using Otsu's method")
    series = series.flatten()
    threshold = threshold_otsu(series)
    labels = (series > threshold).astype(int)

    # evaluate
    logger.debug("Evaluating Otsu thresholding")
    background = series[labels == 0]
    signal = series[labels == 1]
    # Inter-class variance (which Otsu's method maximizes)
    weight1 = np.sum(labels == 0) / len(labels)
    weight2 = np.sum(labels == 1) / len(labels)
    inter_class_variance = (
        weight1 * weight2 * (np.mean(signal) - np.mean(background)) ** 2
    )
    # Calculate entropy of the thresholded image
    hist, _ = np.histogram(labels, bins=2)
    hist_norm = hist / np.sum(hist)
    entropy = scipy.stats.entropy(hist_norm)
    metrics = {
        "inter_class_variance": float(inter_class_variance),
        "entropy": float(entropy),
    }
    return labels, threshold, metrics


def classify_gmm_demux(df, logger=None, **kwargs):
    from hto.external import gmm_demux

    # init
    key_hash = "hash_id"
    htos = df.columns
    logger = logger or get_logger("demux", level=1)

    # demultiplex
    df_demux = gmm_demux(df, hash_id=key_hash, params="--random_seed 42", **kwargs)

    # get meta data
    classifications = {}
    thresholds = {}
    metrics = {}
    for hto in htos:
        # classify
        classifications[hto] = df_demux[key_hash].str.contains(hto).astype(int)

        # estimate threshold
        min_signal = df.loc[classifications[hto] == 1, hto].min()
        max_background = df.loc[classifications[hto] == 0, hto].max()
        thresholds[hto] = (min_signal + max_background) / 2
        logger.debug(f"Threshold for {hto}: {thresholds[hto]}")

        # add to metrics
        metrics[hto] = {
            "min_signal": min_signal,
            "max_background": max_background,
        }

    return classifications, thresholds, metrics



"""Visualise technical noise and HTO expression values."""

from typing import Union

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from .._defaults import DEFAULTS
from .._logging import get_logger


def _plot_layer(
    df,
    axs_row,
    y,
    varname,
    df_line: pd.DataFrame = None,
    hline: float = None,
    **kwargs_fig,
):
    # scatterplot
    sns.scatterplot(
        df,
        x="noise",
        y=y,
        hue="highlight",
        size="highlight",
        ax=axs_row[0],
        **kwargs_fig,
    )
    axs_row[0].set_title(f"{varname} normalised")

    # regression line
    if df_line is not None:
        sns.lineplot(df_line, x="x", y="y", c="black", ax=axs_row[0])

    # KDE plot
    sns.kdeplot(data=df, y="normalised", ax=axs_row[1], fill=True)

    # horizontal line (demultiplexing threshold)
    if hline is not None:
        axs_row[0].axhline(hline, c="grey", linestyle="--")
        axs_row[1].axhline(hline, c="grey", linestyle="--")


def technical_noise(
    adata: ad.AnnData,
    var: Union[int, str] = None,
    add_threshold: bool = True,
    use_key_normalise: str = DEFAULTS["add_key_normalise"],
    use_key_denoise: str = DEFAULTS["add_key_denoised"],
    kwargs_fig: dict = {},
    highlight: np.ndarray = None,
    plot_layers: list = ["normalised", "denoised"],
    axs: plt.Axes = None,
    verbose: int = 1,
):
    """Visualise technical noise in the data.

    Step 1: Create a dataframe with noise, expression values and denoised values.
    Step 2: Create a scatter plot of noise vs expression values.
    Step 3: Create a KDE plot of expression values.
    """
    # assert
    if var is None:
        logger = get_logger("technical_noise", level=verbose)
        logger.error(
            f"Parameter 'var' must be provided. For example, set 'var=0' or 'var='{adata.var_names[0]}''"
        )
        return

    if isinstance(var, int):
        i = var
        varname = adata.var_names[i]
    elif isinstance(var, str):
        assert var in adata.var_names, f"Variable {var} not found in adata.var_names"
        i = np.where(adata.var_names == var)[0][0]
        varname = var
    else:
        raise ValueError("'var' must be either an integer index or a var_name")

    if axs is not None:
        assert len(axs.shape) == 2, "axs must be a 2D numpy array"
        assert axs.shape == (2, 2), "axs must be a 2x2 numpy array"

    # get data
    df = pd.DataFrame(
        {
            "noise": adata.uns["dnd"]["denoise"]["covariates"],
            "normalised": adata.layers[use_key_normalise][:, i],
            "denoised": adata.layers[use_key_denoise][:, i],
        },
        index=adata.obs_names,
    )
    if highlight is not None:
        df["highlight"] = highlight
    else:
        df["highlight"] = 0
    coefs = adata.uns["dnd"]["denoise"]["batch_model"]["coefs"][i]

    # get line
    x = np.array([df.noise.min(), df.noise.max()])
    y = coefs[0] + x * coefs[1]
    df_line = pd.DataFrame({"x": x, "y": y})

    # get thresholds
    threshold_denoised = adata.uns["dnd"]["demux"]["thresholds"][varname]

    # defaults
    defaults = {
        "linewidth": 0,
        "s": 5,
        "alpha": 0.5,
    }
    kwargs_fig = {**defaults, **kwargs_fig}

    # plot
    m = len(plot_layers)
    params_fig = {
        "figsize": (10, m * 5),
        "gridspec_kw": {"width_ratios": [3, 1]},
        "sharey": "row",
    }
    if axs is None:
        fig, axs = plt.subplots(m, 2, squeeze=False, **params_fig)

    # normalised layer
    _plot_layer(
        df=df,
        axs_row=axs[0],
        y="normalised",
        varname=varname,
        df_line=df_line,
        hline=None,
        **kwargs_fig,
    )

    _plot_layer(
        df=df,
        axs_row=axs[1],
        y="denoised",
        varname=varname,
        df_line=None,
        hline=threshold_denoised,
        **kwargs_fig,
    )

    plt.tight_layout()

    return df, axs



import importlib
import os
import sys

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
from matplotlib.backends.backend_pdf import PdfPages
from pandas.api.types import is_float_dtype, is_integer_dtype

from ._defaults import DEFAULTS, DESCRIPTIONS
from ._exceptions import AnnDataFormatError, UserInputError
from ._logging import get_logger
from ._meta import init_meta


def _assert_float(x):
    if not is_float_dtype(x):
        # check if just malformed int
        if scipy.sparse.issparse(x):
            x_test = x.data[:1000]
            x_test = x_test[x_test != 0]
            if np.any(x_test == np.round(x_test)):
                raise AnnDataFormatError("adata_not_float", x_test)
        elif isinstance(x, np.ndarray):
            x_test = x[:1000]
            x_test = x_test[x_test != 0]
            if np.any(x_test == np.round(x_test)):
                raise AnnDataFormatError("adata_not_float", x_test)
        else:
            raise AnnDataFormatError("adata_not_float", x)


def _assert_int(x):
    if not is_integer_dtype(x):
        # check if just malformed float
        if scipy.sparse.issparse(x):
            x_test = x.data[:1000]
            if np.any(x_test != np.round(x_test)):
                raise AnnDataFormatError("adata_not_int", x)
        elif isinstance(x, np.ndarray):
            x_test = x[:1000]
            if np.any(x_test != np.round(x_test)):
                raise AnnDataFormatError("adata_not_int")
        else:
            raise AnnDataFormatError("adata_not_int", x)


def get_layer(
    adata,
    use_layer,
    numpy,
    inplace: bool = False,
    float: bool = False,
    integer: bool = False,
):
    """Get layer from AnnData object and perform checks.

    Args:
        adata (AnnData): AnnData object
        use_layer (str): Layer to use
        float (bool): Assert that the data is float
        numpy (bool): Convert to numpy
        inplace (bool): Return a copy
        integer (bool): Assert that the data is integer

    """
    # copy if not inplace
    if not inplace:
        adata = adata.copy()

    # initialize meta
    adata = init_meta(adata)

    # get layer
    x = adata.X
    if use_layer is not None:
        x = adata.layers[use_layer]

    # to numpy
    if numpy:
        x = to_dense_safe(x)

    # assertions
    if float:
        _assert_float(x)

    if integer:
        _assert_int(x)

    return adata, x


def get_arg(v, kwargs, defaults):
    """Get argument from kwargs or defaults."""
    # if dict, overwrite keys
    if isinstance(defaults[v], dict):
        defaults[v].update(kwargs.get(v, {}))
        return defaults[v]
    return kwargs.get(v, defaults[v])


def subset_whitelist(adata, whitelist, _required_prop=0.9):
    """Subset whitelist, even when not all barcodes are present in the data."""
    # get whitelist
    logger = get_logger("_utils", level=1)
    whitelist = pd.Series(whitelist)
    whitelist_select = whitelist[whitelist.isin(adata.obs.index)]
    whitelist_pct = len(whitelist_select) / len(whitelist) * 100

    # assert
    if whitelist_pct < _required_prop:
        msg = f"Too view of the whitelist barcodes were found in the data: {whitelist_pct:.1f}%"
        msg += f"\n- Example whitelist selected from filtered hto data: {whitelist[:3].tolist()}"
        msg += f"\n- Example background cell ids from raw hto or gex data: {adata.obs_names[:3].tolist()}"
        msg += "\nPlease make sure that whitelist barcodes are present in the data."
        raise UserInputError(msg)
    if whitelist_pct < 1:
        logger.warning(
            f"Some whitelisted barcodes were not found: {whitelist_pct:.1f}%"
        )
    return adata[whitelist_select]


def test_write(path, filetype, create_folder=True, _require_write=False):
    """Test if file can be written before running demultiplexing."""
    # skip
    if path is None:
        return

    # naming
    assert path.endswith(f".{filetype}"), "Output path must end with .h5ad"

    # create folder
    if create_folder:
        os.makedirs(os.path.dirname(path), exist_ok=True)

    # write
    if _require_write:
        with open(path, "w") as f:
            f.write("test")
        os.remove(path)


def _fix_columns(df):
    """Categorical to string type."""
    for k in df.columns:
        if isinstance(df[k], pd.CategoricalDtype):
            df[k] = df[k].astype(str)
    return df


def _fix_layers(adata):
    """All 'matrix' layers must be arrays."""
    if isinstance(adata.X, np.matrix):
        adata.X = np.array(adata.X)
    for layer in adata.layers:
        if isinstance(adata.layers[layer], np.matrix):
            adata.layers[layer] = np.array(adata.layers[layer])
    return adata


def write_h5ad_safe(adata, path, create_folder=True, _require_write=False):
    """Write AnnData object to h5ad file safely.

    Args:
        adata (AnnData): AnnData object
        path (str): Path to save the file
        create_folder (bool): Create folder if it does not exist

    """
    # skip
    if path is None:
        return
    try:
        # create folder
        if create_folder:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        # all categorical columns to str
        adata.obs = _fix_columns(adata.obs)
        # all np.matrix concepts to dense
        adata = _fix_layers(adata)
        # overwrite
        if os.path.exists(path):
            os.remove(path)
        # write
        adata.write_h5ad(path)
    except Exception as e:
        if _require_write:
            raise e
        logger = get_logger("_utils", level=1)
        logger.error(f"Failed to write h5ad file: {path} ('{e}')")


def write_csv_safe(
    adata,
    path,
    key_hashid: str = DEFAULTS["add_key_hashid"],
    key_doublet: str = DEFAULTS["add_key_doublet"],
    create_folder: bool = True,
):
    """Write barcode, hashid, and doublet information to CSV file."""
    if path is None:
        return
    try:
        df = pd.DataFrame(
            {
                "barcode": adata.obs.index,
                key_hashid: adata.obs[key_hashid],
                key_doublet: adata.obs[key_doublet],
            }
        )
        if create_folder:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        df.to_csv(path, index=False)
    except Exception as e:
        logger = get_logger("_utils", level=1)
        logger.error(f"Failed to write csv file: {path} ({e})")


def reload_all(module):
    """Reload a module and all its submodules."""
    module_name = module.__name__

    # Reload submodules
    for submodule in list(sys.modules):
        if submodule.startswith(module_name + "."):
            try:
                importlib.reload(sys.modules[submodule])
            except ModuleNotFoundError:
                print(f"Submodule '{submodule}' not found. Skipping.")

    # Reload the parent module last
    importlib.reload(module)


def savepdf(pdf: PdfPages, path: str, fig: plt.Figure, show: bool = False):
    # open pdf
    if pdf is None and path is not None:
        pdf = PdfPages(path)

    # save
    if pdf is not None:
        pdf.savefig(fig)

    # display
    if not show:
        plt.close()


def _assert_required_inputs(meta, key, kwargs, parameter):
    """Assert that all required inputs are present and not None.

    Args:
        meta (dict): Metadata for the parameter
        key (str): Key to check
        kwargs (dict): Input arguments
        parameter (str): Parameter name used in error message

    Example:
    ```
    meta = {
        "v1": {
            "description": "Description of v1",
            "required": ["a", "b"],
            "optional": ["c"]
        }
    }
    key = "v1"
    kwargs = {
        "a": 1,
        "b": 2,
        "c": 3
    }
    _assert_required_inputs(meta, key, kwargs, parameter="version")
    ```

    """
    # init error message
    error_msg = f"Available options for '{parameter}':\n"
    # add line
    error_msg += "".join(["-"] * 20) + f"\n{parameter}:\n"
    for k, v in meta.items():
        required = ", ".join(v["required"])
        optional = ", ".join(v["optional"])
        error_msg += f"  '{k}'\n"
        error_msg += f"      Description: {v['description']}\n"
        error_msg += f"      Required: {required}\n"
        if len(optional) > 0:
            error_msg += f"      Optional: {optional}\n"

    # throw errors
    if key not in meta:
        msg = f"\nInvalid input: '{key}' for '{parameter}': '{key}'. " + error_msg
        raise UserInputError(msg)
    for var in meta[key]["required"]:
        if var not in kwargs or kwargs[var] is None:
            msg = (
                f"\nMissing required input: '{var}' for '{parameter}': '{key}'. "
                + error_msg
            )
            raise UserInputError(msg)


def user_input_error_decorator(func):
    """Catch UserInputError and print message instead of traceback."""
    logger = get_logger("UserInputError", level=1)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except UserInputError as e:
            logger.error(e)

    return wrapper


def to_dense_safe(x):
    """Convert sparse matrix to dense matrix safely."""
    if scipy.sparse.issparse(x):
        return x.todense()
    return x


def add_docstring():
    def wrapper(func):
        func.__doc__ = func.__doc__.format(**DESCRIPTIONS)
        return func

    return wrapper


@user_input_error_decorator
def read_adata(path):
    if path.endswith(".h5"):
        try:
            import scanpy as sc
        except ImportError:
            raise ImportError(
                "scanpy is not installed. Install with `pip install scanpy`"
            )

        adata = sc.read_10x_h5(path)
    elif path.endswith(".h5ad"):
        adata = ad.read_h5ad(path)
    else:
        raise UserInputError(
            f"Unknown file format for adata: {path}. Must be anndata (.h5ad) or 10x h5 (.h5)"
        )
    return adata


def is_github_actions():
    return os.getenv("GITHUB_ACTIONS") == "true"


def skip_on_github_actions():
    """Skip test on GitHub Actions."""

    def wrapper(func):
        import pytest

        return pytest.mark.skipif(
            is_github_actions(), reason="Skipping on GitHub Actions"
        )(func)

    return wrapper



"""Compare GEX to HTO counts."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def umi_gex_hto(
    adata_hto,
    adata_background,
    adata_hto_raw,
    adata_gex,
    num_buckets=51,
    subsample=10000,
    axs=None,
):
    """Plot expression level for GEX on x-axis and HTO on y-axis."""
    # combine data
    wl = adata_hto.obs_names
    sgex = pd.Series(
        name="gex",
        data=np.array(adata_gex.X.sum(axis=1)).flatten(),
        index=adata_gex.obs_names,
    )
    shto = pd.Series(
        name="hto",
        data=np.array(adata_hto_raw.X.sum(axis=1)).flatten(),
        index=adata_hto_raw.obs_names,
    )
    df = pd.concat([np.log10(sgex + 1), np.log10(shto + 1)], join="inner", axis=1)

    # make and summarise buckets
    cuts = np.linspace(df.gex.min(), df.gex.max(), num_buckets)
    df.loc[:, "gex_bucket"] = pd.cut(df.gex, cuts, labels=cuts[:-1])
    df.loc[:, "filtered"] = df.index.isin(wl)
    df = df[df.gex > np.log10(8)]

    # build background
    df.loc[:, "background"] = df.index.isin(adata_background.obs_names)

    # add labels
    df.loc[:, "label"] = "Empty"
    df.loc[df.background, "label"] = "Ambient Droplets"
    df.loc[df.filtered, "label"] = "Cells"

    # aggregate
    df_agg = (
        df.groupby("gex_bucket", observed=True)
        .agg(
            **{
                "hto": ("hto", "mean"),
                "cell_frac": ("filtered", "mean"),
                "barcodes": ("filtered", "size"),
            }
        )
        .reset_index()
    )

    # plot
    if axs is None:
        fig, axs = plt.subplots(1, 2, figsize=(8, 4), dpi=150)

    ax = axs[0]
    ax_frac = ax.twinx()

    # % cells per gex bucket
    ax_frac = sns.lineplot(
        df_agg,
        x="gex_bucket",
        y="cell_frac",
        c="lightgrey",
        ax=ax_frac,
    )

    ax = sns.scatterplot(
        df.sample(min(subsample, df.shape[0])),
        x="gex",
        y="hto",
        linewidth=0,
        s=5,
        alpha=0.4,
        hue="label",
        ax=ax,
    )

    # mean hto per gex bucket
    ax = sns.lineplot(
        df_agg,
        x="gex_bucket",
        y="hto",
        c="black",
        linewidth=3,
        ax=ax,
    )

    # distribution of hto
    ax = axs[1]
    sns.kdeplot(
        df,
        x="hto",
        hue="label",
        fill=True,
        common_norm=False,
        ax=ax,
    )

    plt.tight_layout()
    return df



"""Methods for demultiplexing HTO data, containing implementation of various binary demultiplexing methods."""

from pprint import pformat

import anndata as ad
import numpy as np
import pandas as pd
import scipy.sparse

from ._classify import assert_demux, classify
from ._defaults import DEFAULTS
from ._exceptions import UserInputError
from ._logging import get_logger
from ._meta import add_meta
from ._utils import add_docstring, get_layer


@add_docstring()
def demux(
    adata_hto: ad.AnnData,
    demux_method: str = DEFAULTS["demux_method"],
    use_layer: str = DEFAULTS["use_layer"],
    key_normalise: str = DEFAULTS["add_key_normalise"],
    enforce_larger_than_background: bool = DEFAULTS["enforce_larger_than_background"],
    add_key_hashid: str = DEFAULTS["add_key_hashid"],
    add_key_doublet: str = DEFAULTS["add_key_doublet"],
    add_key_labels: str = DEFAULTS["add_key_labels"],
    kwargs_classify: dict = DEFAULTS["kwargs_classify"],
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
):
    """Classify HTOs as singlets (assign to HTO), doublets, or negatives.

    Use a 2-component K-means, GMM, or Otsu threshold method to categorize cells based on their HTO classifications.

    Args:
        adata_hto (ad.AnnData): {adata_hto}
        demux_method (str): {demux_method}
        use_layer (str): {use_layer}
        key_normalise (str): {add_key_normalise}
        enforce_larger_than_background (bool): {enforce_larger_than_background}
        add_key_hashid (str): {add_key_hashid}
        add_key_doublet (str): {add_key_doublet}
        add_key_labels (str): {add_key_labels}
        kwargs_classify (dict): {kwargs_classify}
        inplace (bool): {inplace}
        verbose (int): {verbose}

    Returns:
        AnnData: An AnnData object containing the results of the demultiplexing in .obs.

    """
    # debug - print parameters
    logger = get_logger("demux", level=verbose)
    params = {
        k: v.shape if isinstance(v, ad.AnnData) else v for k, v in locals().items()
    }
    params_str = pformat(params, indent=4)
    logger.debug(f"Parameters:\n{params_str}")

    # setup data
    adata_hto, _ = get_layer(
        adata=adata_hto,
        use_layer=use_layer,
        numpy=True,
        float=True,
        inplace=inplace,
    )

    # assertions
    assert_demux(demux_method)
    if add_key_hashid in adata_hto.obs.columns:
        logger.warning(
            f"Column '{add_key_hashid}' already exists in adata.obs. Overwriting and storing previous columns under '{add_key_hashid}_archive'."
        )
        adata_hto.obs[f"{add_key_hashid}_archive"] = adata_hto.obs[add_key_hashid]
        adata_hto.obs.drop(add_key_hashid, axis=1, inplace=True)

    if enforce_larger_than_background:
        if key_normalise not in adata_hto.layers.keys():
            raise UserInputError(
                f"Layer '{key_normalise}' not found in adata. Please normalise the data first or set 'enforce_larger_than_background' to False."
            )

    # get data
    df_umi = adata_hto.to_df(layer=use_layer)
    assert all([np.issubdtype(t, np.floating) for t in df_umi.dtypes]), (
        "Denoised data must be float."
    )

    # Get classifications for each HTO
    logger.debug(f"Starting demultiplexing using '{demux_method}'...")
    classifications, thresholds, metrics = classify(
        data=df_umi,
        demux_method=demux_method,
        verbose=verbose,
        kwargs_classify=kwargs_classify,
    )

    # Enforce that positively labeled cells are larger than background. This can be skewed due to the technical noise correction of small cells.
    if enforce_larger_than_background:
        logger.debug("Enforcing larger than background...")
        df_normalised = adata_hto.to_df(layer=key_normalise)
        for hto in df_normalised.columns:
            classifications[hto][df_normalised[hto] <= 0] = 0

    # Init results
    logger.debug("Assigning labels...")
    labels_df = pd.DataFrame(classifications, index=df_umi.index)
    result_df = pd.DataFrame(
        "",
        index=labels_df.index,
        columns=[add_key_hashid, add_key_doublet],
        dtype="object",
    )

    # Get cell labels
    result_df.loc[:, add_key_hashid] = labels_df.idxmax(axis=1)
    result_df.loc[labels_df.sum(axis=1) == 0, add_key_hashid] = "negative"
    result_df.loc[labels_df.sum(axis=1) >= 2, add_key_hashid] = "doublet"

    # Get doublet info
    logger.debug("Assigning doublet info...")

    def _assign_doublet(row):
        if row.sum() == 0:
            return "negative"
        return ":".join([r for r in row[row == 1].index])

    result_df.loc[:, add_key_doublet] = labels_df.apply(_assign_doublet, axis=1)

    # Append to AnnData
    if add_key_labels is not None:
        adata_hto.layers[add_key_labels] = scipy.sparse.csr_matrix(labels_df.values)
    adata_hto.obs = pd.concat([adata_hto.obs, result_df], axis=1)

    # add metadata
    adata_hto = add_meta(
        adata_hto,
        step="demux",
        params=dict(
            demux_method=demux_method,
        ),
        metrics=metrics,
        thresholds=thresholds,
    )

    pct_doublet = (result_df[add_key_hashid] == "doublet").mean() * 100
    pct_negative = (result_df[add_key_hashid] == "negative").mean() * 100
    logger.info(
        f"Demultiplexing completed ({pct_doublet:.1f}% Doublets, {pct_negative:.1f}% Negatives). Hash IDs stored in '{add_key_hashid}'."
    )
    return adata_hto



"""HTO Plotting Package."""

from .barplot import barplot
from .distributions import distribution, distribution_stages
from .heatmap import heatmap
from .technical_noise import technical_noise
from .umap import umap
from .umi_gex_hto import umi_gex_hto
from .umiplot import umi

__all__ = [
    "distribution",
    "distribution_stages",
    "umi",
    "technical_noise",
    "umi_gex_hto",
    "umap",
    "barplot",
    "heatmap",
]



"""Collection of utility functions for generating mock HTO expression data."""

import anndata as ad
import numpy as np
import pandas as pd
from sklearn.datasets import make_blobs


def generate_hto(n_cells=1000, n_htos=3, noise_level=0.5, seed=42):
    """Generate clustered HTO data.

    Creates the following labels:
    - "filtered"
        - 50% Singlet: High in one HTO cluster
        - 10% Doublet: High in two HTO clusters
        - 40% Negative: Low in all HTO clusters
    - "raw"
        - +100% Empty: Very low in all HTO clusters
    - "gex"
        - 200%: 15 genes, summing to 300 for each cell and below 100 for empty droplets

    Returns:
        Dictionary with the following keys:
        - 'filtered': AnnData object with filtered data
        - 'path_filtered': Path to the filtered data
        - 'raw': Raw data
        - 'path_raw': Path to the raw data
        - 'path_denoised': Path to the output data
        - 'filtered_labels': Labels for the filtered data
        - 'raw_labels': Labels for the raw data
        - 'filtered_cell_types': Cell types for the filtered data
        - 'raw_cell_types': Cell types for the raw data

    """
    # Set seed
    np.random.seed(seed)
    assert n_htos >= 2, "Number of HTOs must be at least 2"

    # Define cluster centers
    # - off-diagonal is random around 0-3
    # - diagonal is random around 10
    hto_centers = np.random.uniform(0, 3, (n_htos, n_htos))
    hto_centers[np.diag_indices(n_htos)] = np.random.uniform(8, 15, n_htos)

    negative_center = np.ones(n_htos) * 2  # All HTOs low
    empty_center = np.ones(n_htos) * 0.5  # All HTOs very low

    # Calculate number of cells for each type
    n_singlets = int(n_cells * 0.5)
    n_doublets = int(n_cells * 0.1)
    n_negatives = n_cells - n_singlets - n_doublets
    n_empty = int(n_cells) * 2  # 100% additional empty droplets used for background

    # Generate singlets
    singlets, singlet_labels = make_blobs(
        n_samples=n_singlets, n_features=n_htos, centers=hto_centers, cluster_std=3.0
    )

    # Generate doublets
    doublets = []
    doublet_labels = []
    for _ in range(n_doublets):
        hto1, hto2 = np.random.choice(n_htos, 2, replace=False)
        doublet = (hto_centers[hto1] + hto_centers[hto2]) * np.random.uniform(0.5, 1)
        doublet += np.random.normal(0, 0.2, n_htos)
        doublets.append(doublet)
        doublet_labels.append("doublet")
    doublets = np.array(doublets)

    # Generate negatives
    negatives, _ = make_blobs(
        n_samples=n_negatives,
        n_features=n_htos,
        centers=[negative_center],
        cluster_std=0.5,
    )

    # Generate empty droplets
    empty, _ = make_blobs(
        n_samples=n_empty, n_features=n_htos, centers=[empty_center], cluster_std=0.2
    )

    # Combine all data
    raw_data = np.vstack((singlets, doublets, negatives, empty))

    # Create labels
    cell_ids = [f"cell{i + 1}" for i in range(raw_data.shape[0])]
    hto_ids = [f"HTO_{i + 1}" for i in range(n_htos)]

    singlet_labels = [f"{hto_ids[i]}" for i in singlet_labels]
    negative_labels = ["negative"] * n_negatives
    empty_labels = ["empty"] * n_empty
    raw_labels = np.array(
        singlet_labels + doublet_labels + negative_labels + empty_labels
    )
    raw_cell_types = np.array(
        ["singlet"] * n_singlets + doublet_labels + negative_labels + empty_labels
    )
    subset_filterd = raw_cell_types != "empty"

    # Ensure non-negative values
    raw_data = np.maximum(raw_data, 0)

    # Add non-negative noise
    raw_data += np.random.lognormal(0, noise_level, raw_data.shape)

    # To integer
    raw_data = np.round(raw_data).astype(int)

    # Create anndata
    obs = pd.DataFrame(
        {
            "cell_id": cell_ids,
            "cell_type": raw_cell_types,
            "ground_truth": raw_labels,
        },
        index=cell_ids,
    )

    var = pd.DataFrame(None, index=hto_ids)

    raw_adata = ad.AnnData(
        X=raw_data,
        dtype="int32",
        obs=obs,
        var=var,
    )
    filtered_adata = raw_adata[subset_filterd].copy()

    # Create GEX data (15 genes)
    # -> Sum is over 300 for each cell
    # -> Sum is below 100 for empty droplets
    counts_cells = np.random.randint(290, 1000, 2 * n_cells)
    counts_empty = np.random.randint(50, 100, n_cells)
    counts = np.concatenate([counts_cells, counts_empty])
    p_distribute = np.random.uniform(0, 1, 15)
    p_distribute /= p_distribute.sum()
    # Distribute
    gex = np.matmul(counts.reshape(-1, 1), p_distribute.reshape(1, -1))
    assert np.all(np.round(gex.sum(axis=1)).astype(int) == counts)
    gex = np.round(gex).astype(int)
    gex_adata = ad.AnnData(
        X=gex,
        dtype="float32",
        obs=obs,
    )

    # return
    return {
        "filtered": filtered_adata,
        "raw": raw_adata,
        "gex": gex_adata,
        "filtered_labels": raw_labels[subset_filterd],
        "raw_labels": raw_labels,
        "filtered_cell_types": raw_cell_types[subset_filterd],
        "raw_cell_types": raw_cell_types,
    }
